package org.bdware.sc.units;

import com.google.gson.JsonArray;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.bdware.sc.bean.ContractExecType;
import org.bdware.sc.bean.ContractRequest;
import org.bdware.sc.bean.IDSerializable;
import org.bdware.sc.conn.ResultCallback;
import org.bdware.sc.db.CMTables;
import org.bdware.sc.db.KeyValueDBUtil;
import org.bdware.sc.redo.TransRecord;
import org.bdware.sc.util.JsonUtil;
import org.bdware.server.trustedmodel.ContractUnitStatus;

import java.util.Map;
import java.util.PriorityQueue;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicInteger;

// 在每个节点上均有
public class MultiContractMeta implements IDSerializable {
    private static final Logger LOGGER = LogManager.getLogger(MultiContractMeta.class);
    private final AtomicInteger lastExeSeq; // last executed request seq
    public volatile int curExeSeq = -1; // 当前正在执行请求序号   for multipoint contract requests
    public String memory;
    public String key; // privateKey
    public String publicKey;
    public String invokeID; // TODO
    public ContractExecType type;
    public ContractUnitStatus unitStatus = ContractUnitStatus.CommonMode;
    public transient PriorityQueue<ContractRequest> queue; // contract request
    public transient Map<Integer, String> uniReqIDMap; // 用于请求
    public transient Map<Integer, ResultCallback> resultMap; // 用于请求
    public transient PriorityQueue<TransRecord> transQueue; // transRecord
    AtomicInteger masterOrder = new AtomicInteger(-1);
    private String contractID;
    private boolean isPrivate = false;
    private String pubKeyPath;
    private String ypkName;
    private String[] members; // 执行这个合约的所有节点的pubKey
    private boolean isMaster;
    private String masterNode;

    public MultiContractMeta(String s) {
        contractID = s;
        isMaster = false;
        masterNode = null;
        lastExeSeq = new AtomicInteger(-1);
        initQueue();
    }

    public void initQueue() {
        queue = new PriorityQueue<>();
        transQueue = new PriorityQueue<>();
        uniReqIDMap = new ConcurrentHashMap<>();
        resultMap = new ConcurrentHashMap<>();
    }

    public int getCurSeqAtMaster() {
        return masterOrder.get();
    }

    public void nextSeqAtMaster() {
        masterOrder.getAndIncrement();
    }

    public void setSeqAtMaster(int seq) {
        masterOrder = new AtomicInteger(seq);
    }

    public void init(MultiContractMeta old) {
        if (old != null && old.hasQueues()) {
            queue = old.queue;
            transQueue = old.transQueue;
            uniReqIDMap = old.uniReqIDMap;
            resultMap = old.resultMap;
        } else {
            if (null == queue) {
                queue = new PriorityQueue<>();
            }
            if (null == transQueue) {
                transQueue = new PriorityQueue<>();
            }
            if (null == uniReqIDMap) {
                uniReqIDMap = new ConcurrentHashMap<>();
            }
            if (null == resultMap) {
                resultMap = new ConcurrentHashMap<>();
            }
        }
    }

    private boolean hasQueues() {
        return null != queue && null != transQueue && null != uniReqIDMap && null != resultMap;
    }

    public int getLastExeSeq() {
        return this.lastExeSeq.get();
    }

    public void setLastExeSeq(int lastExeSeq) {
        this.lastExeSeq.set(lastExeSeq);
        if (KeyValueDBUtil.instance.containsKey(
                CMTables.LastExeSeq.toString(), contractID)) { // 如果现在是Stable模式就同步刷到磁盘
            KeyValueDBUtil.instance.setValue(
                    CMTables.LastExeSeq.toString(), contractID, String.valueOf(lastExeSeq));
        }
    }

    public boolean isSequent(int a) {
        return a - lastExeSeq.get() == 1;
    }

    public String getContractID() {
        return contractID;
    }

    public void setContractID(String id) {
        contractID = id;
    }

    public String getID() {
        return contractID;
    }

    public void setIsPrivate(boolean t) {
        isPrivate = t;
    }

    public boolean isPrivate() {
        return isPrivate;
    }

    public String getPubKeyPath() {
        if (!isPrivate) {
            return null;
        }
        return pubKeyPath;
    }

    public void setPubKeyPath(String p) {
        pubKeyPath = p;
    }

    public String getYpkName() {
        return ypkName;
    }

    public void setYpkName(String s) {
        ypkName = s;
    }

    public String[] getMembers() {
        return members;
    }

    public void setMembers(String[] m) {
        members = m;
    }

    public void setMembers(JsonArray members) {
        String[] copied = new String[members.size()];
        for (int i = 0; i < members.size(); i++) {
            copied[i] = members.get(i).getAsString();
        }
        this.members = copied;
    }

    public void printSet() {
        for (ContractRequest cr : queue) {
            LOGGER.debug("请求-" + cr.seq);
        }
    }

    public void addRequestQueue(ContractRequest request, String uniReqID, ResultCallback result) {
        LOGGER.info("put to queue, seq:" + request.seq + " queue.size:" + queue.size());
        queue.add(request);
        int seq = request.seq;
        uniReqIDMap.put(seq, uniReqID);
        resultMap.put(seq, result);
    }

    public void printContent() {
        System.out.println("contractID=" + (contractID == null ? "null" : contractID));
        System.out.println("isPrivate=" + isPrivate);
        System.out.println("pubKeyPath=" + (pubKeyPath == null ? "null" : pubKeyPath));
        System.out.println("ypkName=" + (ypkName == null ? "null" : ypkName));
        System.out.println("key=" + (key == null ? "null" : key));
        System.out.println("publicKey=" + (publicKey == null ? "null" : publicKey));
        System.out.println("invokeID=" + (invokeID == null ? "null" : invokeID));
        System.out.println("type=" + (type == null ? "null" : type));
        System.out.println("lastExeSeq=" + lastExeSeq);
        System.out.println("members=" + (members == null ? "null" : JsonUtil.toJson(members)));
        System.out.println("memory=" + (memory == null ? "null" : memory));
    }

    public void setIsMaster(boolean isMaster) {
        this.isMaster = isMaster;
    }

    public boolean isMaster() {
        return isMaster;
    }

    public void setMaster(String master) {
        masterNode = master;
    }

    public String getMasterNode() {
        return masterNode;
    }

}
